# PRG - Exam #

This is the assignment for the PRG exam. Follow the instructions precisely and please note:

- If the module executes any functions when imported (e.g. print), points will be subtracted.
- If the module fails to be imported (due to syntactic errors), the solution is invalid.

The modules (packages) you implement should be packed in a zip file (files should be in the root of
the zip file, zipping directly the files, not the containing folder) and uploaded to BRUTE. For
every module, class and function you implement, do not forget to add a docstring. Code quality
(mainly following naming conventions and writing docstrings) is part of the grading of your
solution.


## Task 1 [3p] ##

Create a module named "traceback_code" which will produce the following traceback when executed.

```
Traceback (most recent call last):
  File "traceback_code.py", line 8, in <module>
    div(2, 0)
  File "traceback_code.py", line 5, in div
    return a / b
ZeroDivisionError: division by zero
```


## Task 2 [6p] ##

Create a module named "task2_functions" and implement three functions according to the following
unit tests. You can copy the unit tests to the same module.

```
import unittest

class TestFunctions(unittest.TestCase):

    def test_remove_punctuation(self):
        # [1p]
        self.assertEqual("Who", remove_punctuation("!!!Who!?!?!"))
        self.assertEqual("when", "when")
        self.assertEqual("Why", remove_punctuation("===>Why?????"))

    def test_have_common_word(self):
        # [2.5p]
        self.assertTrue(have_common_word(["dog", "cat"], ["cat", "bird"]))
        self.assertTrue(have_common_word(["dog", "cat"], ["CAT", "BIRD"]))
        self.assertTrue(have_common_word(["DOG", "CAT"], ["cat", "bird"]))
        self.assertFalse(have_common_word(["dog", "cat"], ["bird", "fish"]))
        self.assertFalse(have_common_word(["dog", "cat"], []))

    def test_longest_word(self):
        # [2.5p]
        self.assertEqual("weather", longest_word("The weather is nice today"))
        self.assertEqual("like", longest_word("I like good code"))
        self.assertEqual("", longest_word(""))
```


## Task 3 [1p] ##

Create a package with a single module containing the class definition below.

1. [1p] Choose the name of the package and module appropriately. Be sure to follow best practices.

```
class Cryptocurrency:
    """Owning a number of units (float) of a specific cryptocurrency (str) at a price (float)"""

    def __init__(self, amount, currency, unit_price):
        self.amount = amount
        self.currency = currency
        self.unit_price = unit_price

    def __add__(self, other):
        if self.currency != other.currency:
            raise ValueError("Cannot add units for different currencies")
        if self.unit_price != other.unit_price:
            raise ValueError("Cannot add currencies with different unit prices")
        return Cryptocurrency(self.amount + other.amount, self.currency, self.unit_price)

    def __repr__(self):
        return f"{self.amount} {self.currency} @ {self.unit_price}"
```


## Task 3.1 [5p] ##

In the module created in Task 3, implement the following tasks.

1. [2p] Add support for comparing instances using the equality operator "==". Consider the amounts
    to be equal as long as the absolute difference between amounts is less than 1e-6 (0.000001).
2. [3p] Write a unit test for the addition operator "+" and put it in a test case named
    "TestCryptocurrency".

Cryptocurrency example

```
>>> Cryptocurrency(0.1, "BTC", 36896.10) == Cryptocurrency(0.01, "BTC", 36896.10) + Cryptocurrency(0.09, "BTC", 36896.10)
True
```


## Task 3.2 [4p] ##

In the module created in Task 3, implement the following tasks.

1. [2p] Write function "save_crypto" which takes two arguments - a Cryptocurrency instance and
    a file path, and will store the data from the instance under the path. The format of the file is
    up to you.
2. [2p] Write function "load_crypto" which takes a single argument - a file path, and returns
    a Cryptocurrency instance containing the data previously saved by save_crypto.


## Task 3.3 [6p] ##

In the module created in Task 3, implement the following tasks.

1. [1p] Implement class UnitPriceChange that stores a float representing the change of unit price.
2. [1p] Add support for adding UnitPriceChange to a Cryptocurrency instance.
3. [1p] Implement a @property "net_worth" that will compute the price of all units in a single
    Cryptocurrency instance.
4. [3p] Implement class IdealCryptocurrency which inherits from Cryptocurrency and ensures that it
    never looses value, i.e. each UnitPriceChange added to the Cryptocurrency is positive. Ensure
    this by raising an appropriate exception if not. Try to reuse as much code from Cryptocurrency
    as possible.

Cryptocurrency example

```
>>> Cryptocurrency(0.1, "BTC", 36896.10) == Cryptocurrency(0.1, "BTC", 36972.50) + UnitPriceChange(-76.40)
True
```


## Epilogue ##

At the end of the exam, the following files and folders should be present in your zip file. They
should be in the root of the zip file, not in a nested folder.

- traceback_code.py - task 1
- task2_functions.py - task 2
- [folder of chosen name] - task 3
